/***************************************************************************/
/* 			     		    	                           */
/*  Extraction of Centipede Model Parameters            		   */
/* 			     ( Optimal Parameters )		    	   */
/* 			     		    	                           */
/*  B. Kurt, H.T. Demiral, M. Gokmen, "Ayrita Dayali Etkin Bir Goruntu     */
/*      Gosterimi," Proceedings of 6th Conference on Signal Processing and */
/*      Applications, SIU'98, Vol. I, pp. 259-264, Ankara, 1998.           */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Centipede Class Library          	                                   */
/* 			     		    	                           */
/*  Centipede         	     		    	                           */
/*      |              	     		    	                           */
/*    edge              	     	    	                           */
/*      |              	     		    	                           */
/*    image              	     	    	                           */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Includes : Centipede_def.h, trace.h, extract.h, perceptual.h           */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Last modified by    		    	                           */
/*                   Binnur KURT            	                           */
/*  When        	     		    	                           */
/* 		     November 1st 1999    	                           */
/* 			     		    	                           */
/*  Copyleft (C) 1999, the CVIP Lab.                                       */
/* 			     		    	                           */
/*  Version          2.1   		    	                           */
/* 			     		    	                           */
/***************************************************************************/
#include <float>
#include <math>

#ifndef __Centipede__
#define __Centipede__

#include "edge.h"

#include "Centipede_def.h"


class Centipede:public Edge{
      public:
            Centipede()                                 ;
           ~Centipede()                                 ;
      /* Perceptual Grouping                           */
      void  PerceptualGrouping(float,float,float,int)   ;
      /* Contour Tracing                               */
      void  BlobColoring()                              ;
      void  GetNormalVect()                             ;
      void  ExtractOptimalCentipedeModelParameters()    ;
      void  ExtractContourDescriptor()                  ;
      void  GetTheDistance()                            ;
      /* Used to pass model parameters to other objects */
      void  SaveModelToMatrices(unsigned char **,
                                  unsigned char **,
                                  bool)     ;
      /* Compression                                   */
      void  CompressCentipede(char *)                   ;
      /* Decode                                         */
      void  DecodeCentipede(char *)                     ;
      /* Functions used to visualize the model          */
      void  DrawBoundingBox(TImage *)                   ;
      void  SaveContoursToTImage(TImage *,TImage *)     ;
      void  SaveDescriptorToTImage(TImage *)            ;
      void  SaveModelToTImage(TImage *,TImage *,bool)   ;
      void  WriteContourDescriptorToFile(char *)        ;

      /* Object variables                               */
      int **BlobColor                                   ;
      int   MaximumContourLength                        ;
      int   MaximumContrast                             ;
      float MaximumCurvature                            ;
      int   NumberOfContours                            ;
      int   NumberOfEdges                               ;
      ContourDescriptorStruct *ContourDescriptor        ;
      float LambdaForData                               ;
      void init();
      /* Operator overriding                           */
      friend CoordinateStruct operator+(CoordinateStruct,CoordinateStruct)     ;
      friend CoordinateStruct operator-(CoordinateStruct,CoordinateStruct)     ;

      private:
      CoordinateStruct Displacement[8]                  ;

      float Fn[66]                                      ;
      float Gn[66]                                      ;
      int   X[3]                                        ;
      float d[4]                                        ;
      float G[4][4]                                     ;
      float c[4]                                        ;
      float optimal_c[4]                                ;
      float ErrorSurface[32][32]                        ;
      /* For coding purposes                            */
      char  HUFFMAN_CHAIN_CODE[8]                       ;
      unsigned int CHAIN_CODE_FRQ[8]                    ;
      void  GetTheChainCode(void)                       ;
      void  GetTheDifferentialChainCode(void)           ;
      void  GetTheStatistics(void)                      ;
      int   Uncompress(int)                             ;

      short int distLeft,distRight                      ;
      void GetTheDataToBeModelled(int,int)              ;
      void ApplyMedianFilterToTheData(int)              ;
      void SearchForTheOptimalModelParameters(int,int)  ;
      void Solve(int)                                   ;

      bool HasBlobColoringBeenDone                      ;
      bool HasContourDescriptorExtracted                ;
      CoordinateStruct GetNextEndPoint()                ;
      CoordinateStruct GetFirstPoint()                  ;
      void FollowEndPoint(CoordinateStruct,int)         ;
      void GetContourInfo(int)                          ;
      void FollowThePoint(CoordinateStruct,int,int,int) ;
      CoordinateStruct FindEndPointOfTheContour(int)    ;
      CoordinateStruct FindFirstPointOfTheContour(int)  ;
} ;

/********************************************************/
/* Here comes the implementations                       */
/********************************************************/
#include "perceptual.h"
#include "trace.h"
#include "extract.h"
#include "coding.h"
#include "decoding.h"

     Centipede::Centipede(){

      NumberOfContours = 0                              ;
      NumberOfEdges    = 0                              ;
      HasBlobColoringBeenDone = false                   ;
      HasContourDescriptorExtracted = false             ;

      Displacement[0].x=0; Displacement[0].y=1          ;
      Displacement[1].x=-1;Displacement[1].y=1          ;
      Displacement[2].x=-1;Displacement[2].y=0          ;
      Displacement[3].x=-1;Displacement[3].y=-1         ;
      Displacement[4].x=0 ;Displacement[4].y=-1         ;
      Displacement[5].x=1 ;Displacement[5].y=-1         ;
      Displacement[6].x=1 ;Displacement[6].y=0          ;
      Displacement[7].x=1 ;Displacement[7].y=1          ;
}


     Centipede::~Centipede(){
      int k                                     ;

      if(HasContourDescriptorExtracted == true){
        for(k=0;k<NumberOfContours;k++)
           free(ContourDescriptor[k].ContourList) ;
        free(ContourDescriptor)                         ;
        delete[] BlobColor[0];
        delete[] BlobColor;
      }else
        if (HasBlobColoringBeenDone == true){
           delete[] BlobColor[0];
           delete[] BlobColor;
        }
}

void Centipede::init(){
      int k                                     ;
      FreeTemporary();
      if(HasContourDescriptorExtracted == true){
        for(k=0;k<NumberOfContours;k++)
           free(ContourDescriptor[k].ContourList) ;
        free(ContourDescriptor)                         ;
        delete[] BlobColor[0];
        delete[] BlobColor;
      }else
        if (HasBlobColoringBeenDone == true){
           delete[] BlobColor[0];
           delete[] BlobColor;
        }

      NumberOfContours = 0                              ;
      NumberOfEdges    = 0                              ;
      HasBlobColoringBeenDone = false                   ;
      HasContourDescriptorExtracted = false             ;

      Displacement[0].x=0; Displacement[0].y=1          ;
      Displacement[1].x=-1;Displacement[1].y=1          ;
      Displacement[2].x=-1;Displacement[2].y=0          ;
      Displacement[3].x=-1;Displacement[3].y=-1         ;
      Displacement[4].x=0 ;Displacement[4].y=-1         ;
      Displacement[5].x=1 ;Displacement[5].y=-1         ;
      Displacement[6].x=1 ;Displacement[6].y=0          ;
      Displacement[7].x=1 ;Displacement[7].y=1          ;
}

CoordinateStruct operator+(CoordinateStruct a,CoordinateStruct b){
          CoordinateStruct c       ;

          c.x = a.x + b.x       ;
          c.y = a.y + b.y       ;
          return c      ;
}

CoordinateStruct operator-(CoordinateStruct a,CoordinateStruct b){
          CoordinateStruct c       ;

          c.x = a.x - b.x       ;
          c.y = a.y - b.y       ;
          return c      ;
}

void Centipede::WriteContourDescriptorToFile(char *filename){
     FILE *hFile        ;
     int  k             ;
     hFile = fopen(filename,"w+")       ;
     if(hFile!=NULL)
     for(k=0;k<ContourDescriptor[k].ContourLength;k++)
        if(ContourDescriptor[k].IsActive == true)
        fwrite(ContourDescriptor,sizeof(ContourDescriptorStruct),
                             NumberOfContours,hFile) ;
     fclose(hFile)      ;
}

void Centipede::SaveContoursToTImage(TImage *img1,TImage *img2){
     int i,j,k,l;
     TRect r;
     int *p,cl;
     unsigned char *q;

     img1->Hint = "BlobColor[][]"                                    ;
     img2->Left = img1->Width + 18                                   ;
     img2->Width = img1->Width                                       ;
     img2->Height = img1->Height                                     ;
     img2->Hint = "ContourDescriptor[]"                              ;

     for(k=0;k<NumberOfContours;k++)
        if(ContourDescriptor[k].IsActive == true)
        for(l=0;l<ContourDescriptor[k].ContourLength;l++){
            i = ContourDescriptor[k].ContourList[l].location.x  ;
            j = ContourDescriptor[k].ContourList[l].location.y  ;
            img1->Canvas->Pixels[j][i] = (TColor) ((k+1) * 6789) ;
        }

     img1->Refresh();
     img2->Refresh();
}

void Centipede::SaveDescriptorToTImage(TImage *img){
     int i,j,k,l                                                 ;


     img->Hint   = "Contour Descriptor"                          ;

     img->Width  = Width                                         ;

     img->Height = Height                                        ;


     for(k=0;k<NumberOfContours;k++)
        if(ContourDescriptor[k].IsActive == true)
        for(l=0;l<ContourDescriptor[k].ContourLength;l++){
            i = ContourDescriptor[k].ContourList[l].location.x  ;
            j = ContourDescriptor[k].ContourList[l].location.y  ;
            img->Canvas->Pixels[j][i] = (TColor) ((k+1) * 6789) ;
     }
}

void  Centipede::DrawBoundingBox(TImage *img){
      int k     ;
      TRect myRect;

      img->Canvas->Brush->Style = bsSolid       ;
      img->Hint = "Bounding Box"                ;

      for(k=0;k<NumberOfContours;k++){
        if(ContourDescriptor[k].IsActive == true){
          img->Canvas->Brush->Color = (TColor) ((k+1) * 6789)      ;
          myRect.Top = ContourDescriptor[k].BoundingBox.BoxBeginV-1;
    	  myRect.Left = ContourDescriptor[k].BoundingBox.BoxBeginH-1;
    	  myRect.Bottom = ContourDescriptor[k].BoundingBox.BoxBeginV+
                          ContourDescriptor[k].BoundingBox.BoxHeight+1;
          myRect.Right =  ContourDescriptor[k].BoundingBox.BoxBeginH+
                          ContourDescriptor[k].BoundingBox.BoxWidth+1;
    	  img->Canvas->FrameRect(myRect);
        }
      }
}
#endif

